"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.initializeInboxListener = void 0;
const error_1 = require("../common/error");
const utils_1 = require("../common/utils");
const logical_replication_listener_1 = require("../replication/logical-replication-listener");
const concurrency_strategy_1 = require("../strategies/concurrency-strategy");
const listener_restart_strategy_1 = require("../strategies/listener-restart-strategy");
const message_processing_db_client_strategy_1 = require("../strategies/message-processing-db-client-strategy");
const message_processing_timeout_strategy_1 = require("../strategies/message-processing-timeout-strategy");
const message_processing_transaction_level_strategy_1 = require("../strategies/message-processing-transaction-level-strategy");
const message_retry_strategy_1 = require("../strategies/message-retry-strategy");
const poisonous_message_retry_strategy_1 = require("../strategies/poisonous-message-retry-strategy");
const inbox_message_storage_1 = require("./inbox-message-storage");
/**
 * Initialize the listener to watch for inbox table inserts.
 * @param config The configuration object with required values to connect to the WAL.
 * @param messageHandlers A list of message handlers to handle the inbox messages.
 * @param logger A logger instance for logging trace up to error logs
 * @param strategies Strategies to provide custom logic for handling specific scenarios
 * @returns Functions for a clean shutdown.
 */
const initializeInboxListener = (config, messageHandlers, logger, strategies) => {
    const allStrategies = applyDefaultStrategies(strategies, config, logger);
    const messageHandlerDict = getMessageHandlerDict(messageHandlers);
    const messageHandler = createMessageHandler(messageHandlerDict, allStrategies, config, logger);
    const errorResolver = createErrorResolver(messageHandlerDict, allStrategies, config, logger);
    const [shutdown] = (0, logical_replication_listener_1.createLogicalReplicationListener)(config, messageHandler, errorResolver, logger, allStrategies, 'inbox');
    return [
        () => __awaiter(void 0, void 0, void 0, function* () {
            var _a;
            yield Promise.all([
                (_a = allStrategies.messageProcessingDbClientStrategy) === null || _a === void 0 ? void 0 : _a.shutdown(),
                shutdown(),
            ]);
        }),
    ];
};
exports.initializeInboxListener = initializeInboxListener;
/**
 * Executes the inbox verification, the actual message handler, and marks the
 * inbox message as processed in one transaction.
 */
const createMessageHandler = (messageHandlers, strategies, config, logger) => {
    return (message) => __awaiter(void 0, void 0, void 0, function* () {
        const transactionLevel = strategies.messageProcessingTransactionLevelStrategy(message);
        if (config.settings.enablePoisonousMessageProtection !== false) {
            const attempt = yield (0, utils_1.executeTransaction)(yield strategies.messageProcessingDbClientStrategy.getClient(message), (client) => __awaiter(void 0, void 0, void 0, function* () {
                // Increment the started_attempts
                const result = yield (0, inbox_message_storage_1.startedAttemptsIncrement)(message, client, config);
                if (result !== true) {
                    logger.warn(message, `Received inbox message cannot have the started attempts incremented: ${result}`);
                    yield client.query('ROLLBACK'); // don't increment the start attempts again on a processed message
                    return false;
                }
                // The startedAttempts was increase in `startedAttemptsIncrement` so the difference is always at least one
                const diff = message.startedAttempts - message.finishedAttempts;
                if (diff >= 2) {
                    const retry = strategies.poisonousMessageRetryStrategy(message);
                    if (!retry) {
                        const msg = `Stopped processing the message with ID ${message.id} as it is likely a poisonous message.`;
                        logger.error(new error_1.TransactionalOutboxInboxError(msg, 'POISONOUS_MESSAGE'), msg);
                        return false;
                    }
                }
                return true;
            }), transactionLevel);
            if (!attempt) {
                return;
            }
        }
        yield (0, utils_1.executeTransaction)(yield strategies.messageProcessingDbClientStrategy.getClient(message), (client) => __awaiter(void 0, void 0, void 0, function* () {
            // lock the inbox message from further processing
            const result = yield (0, inbox_message_storage_1.initiateInboxMessageProcessing)(message, client, config);
            if (result !== true) {
                logger.warn(message, `Received inbox message cannot be processed: ${result}`);
                return;
            }
            // Execute the message handler
            const handler = messageHandlers[getMessageHandlerKey(message)];
            if (handler) {
                yield handler.handle(message, client);
            }
            else {
                logger.debug(`No message handler found for aggregate type "${message.aggregateType}" and message tye "${message.messageType}"`);
            }
            yield (0, inbox_message_storage_1.markInboxMessageCompleted)(message, client, config);
        }), transactionLevel);
    });
};
const getMessageHandlerKey = (h) => `${h.aggregateType}@${h.messageType}`;
const getMessageHandlerDict = (messageHandlers) => {
    const handlers = {};
    for (const handler of messageHandlers) {
        const key = getMessageHandlerKey(handler);
        if (key in handlers) {
            throw new error_1.TransactionalOutboxInboxError(`Only one message handler can handle one aggregate and message type. Multiple message handlers try to handle the aggregate type "${handler.aggregateType}" with the message type "${handler.messageType}".`, 'CONFLICTING_MESSAGE_HANDLERS');
        }
        handlers[key] = handler;
    }
    if (Object.keys(handlers).length === 0) {
        throw new error_1.TransactionalOutboxInboxError('At least one message handler must be provided.', 'NO_MESSAGE_HANDLER_REGISTERED');
    }
    return handlers;
};
/**
 * Increase the "finished_attempts" counter of the message and (potentially) retry it later.
 */
const createErrorResolver = (messageHandlers, strategies, config, logger) => {
    /**
     * An error handler that will increase the "finished_attempts" counter on
     * transient errors or sets the "finished_attempts" counter to max attempts
     * for permanent errors. The message handler `handleError` is called to allow
     * your code to decide if the error is a transient or a permanent one.
     * @param message: the InboxMessage that failed to be processed
     * @param error: the error that was thrown while processing the message
     */
    return (message, error) => __awaiter(void 0, void 0, void 0, function* () {
        const handler = messageHandlers[getMessageHandlerKey(message)];
        const transactionLevel = strategies.messageProcessingTransactionLevelStrategy(message);
        try {
            return yield (0, utils_1.executeTransaction)(yield strategies.messageProcessingDbClientStrategy.getClient(message), (client) => __awaiter(void 0, void 0, void 0, function* () {
                message.finishedAttempts++;
                const shouldRetry = strategies.messageRetryStrategy(message);
                if (handler === null || handler === void 0 ? void 0 : handler.handleError) {
                    yield handler.handleError(error, message, client, shouldRetry);
                }
                if (shouldRetry) {
                    logger.warn(error, `An error ocurred while processing the message with id ${message.id}.`);
                }
                else {
                    logger.error((0, error_1.ensureExtendedError)(error, 'GIVING_UP_MESSAGE_HANDLING'), `Giving up processing the message with id ${message.id}.`);
                }
                yield (0, inbox_message_storage_1.increaseInboxMessageFinishedAttempts)(message, client, config);
                return shouldRetry;
            }), transactionLevel);
        }
        catch (error) {
            const msg = handler
                ? 'The error handling of the message failed. Please make sure that your error handling code does not throw an error!'
                : 'The error handling of the message failed.';
            logger.error(new error_1.MessageError(msg, 'MESSAGE_HANDLING_FAILED', message, error), msg);
            // In case the error handling logic failed do a best effort to increase the "finished_attempts" counter
            try {
                yield (0, utils_1.executeTransaction)(yield strategies.messageProcessingDbClientStrategy.getClient(message), (client) => __awaiter(void 0, void 0, void 0, function* () {
                    yield (0, inbox_message_storage_1.increaseInboxMessageFinishedAttempts)(message, client, config);
                }), transactionLevel);
                return true;
            }
            catch (_a) {
                return false;
            }
        }
    });
};
const applyDefaultStrategies = (strategies, config, logger) => {
    var _a, _b, _c, _d, _e, _f, _g;
    return ({
        concurrencyStrategy: (_a = strategies === null || strategies === void 0 ? void 0 : strategies.concurrencyStrategy) !== null && _a !== void 0 ? _a : (0, concurrency_strategy_1.defaultConcurrencyStrategy)(),
        messageProcessingDbClientStrategy: (_b = strategies === null || strategies === void 0 ? void 0 : strategies.messageProcessingDbClientStrategy) !== null && _b !== void 0 ? _b : (0, message_processing_db_client_strategy_1.defaultMessageProcessingDbClientStrategy)(config, logger),
        messageProcessingTimeoutStrategy: (_c = strategies === null || strategies === void 0 ? void 0 : strategies.messageProcessingTimeoutStrategy) !== null && _c !== void 0 ? _c : (0, message_processing_timeout_strategy_1.defaultMessageProcessingTimeoutStrategy)(config),
        messageProcessingTransactionLevelStrategy: (_d = strategies === null || strategies === void 0 ? void 0 : strategies.messageProcessingTransactionLevelStrategy) !== null && _d !== void 0 ? _d : (0, message_processing_transaction_level_strategy_1.defaultMessageProcessingTransactionLevelStrategy)(),
        messageRetryStrategy: (_e = strategies === null || strategies === void 0 ? void 0 : strategies.messageRetryStrategy) !== null && _e !== void 0 ? _e : (0, message_retry_strategy_1.defaultMessageRetryStrategy)(config),
        poisonousMessageRetryStrategy: (_f = strategies === null || strategies === void 0 ? void 0 : strategies.poisonousMessageRetryStrategy) !== null && _f !== void 0 ? _f : (0, poisonous_message_retry_strategy_1.defaultPoisonousMessageRetryStrategy)(config),
        listenerRestartStrategy: (_g = strategies === null || strategies === void 0 ? void 0 : strategies.listenerRestartStrategy) !== null && _g !== void 0 ? _g : (0, listener_restart_strategy_1.defaultListenerRestartStrategy)(config),
    });
};
//# sourceMappingURL=inbox-listener.js.map