import { ExtendedError } from '../common/error';
import { TransactionalLogger } from '../common/logger';
import { TransactionalOutboxInboxConfig } from '../replication/config';
import { ListenerType } from '../replication/logical-replication-listener';
/**
 * When some error is caught in the inbox or outbox listener this strategy will
 * allow you to log/track the error and potentially remove the underlying issue
 * e.g. if the replication slot does not exist after a database failover. It
 * returns the wait time until the listener should try to connect again.
 */
export interface ListenerRestartStrategy {
    /**
     * Check based on the error how long the listener should wait to restart.
     * @param error The caught error object (non Error type errors are wrapped)
     * @param logger The logger instance to use for logging the error.
     * @param listenerType The inbox or outbox listener name
     * @returns The time in milliseconds how long the listener should wait before restarting
     */
    (error: ExtendedError, logger: TransactionalLogger, listenerType: ListenerType): Promise<number>;
}
/**
 * The default listener restart strategy checks if the error is a PostgreSQL
 * error. If the PostgreSQL error is about the replication slot being in use, it
 * logs a trace entry and waits for the configured `restartDelaySlotInUse`
 * (default: 10sec) time. Otherwise, it logs an error entry and waits for the
 * configured `restartDelay` (default: 250ms).
 */
export declare const defaultListenerRestartStrategy: (config: TransactionalOutboxInboxConfig) => ListenerRestartStrategy;
/**
 * The default listener and slot restart strategy uses the same logic as the
 * `defaultListenerRestartStrategy`. In addition, it checks if a PostgreSQL error
 * is about the replication slot not existing (e.g. after a DB failover). Then
 * it tries to create the replication slot with the connection details of the
 * replication user slot and waits for the configured `restartDelay` (default:
 * 250ms).
 */
export declare const defaultListenerAndSlotRestartStrategy: (config: TransactionalOutboxInboxConfig) => ListenerRestartStrategy;
//# sourceMappingURL=listener-restart-strategy.d.ts.map